/*
 *    fwdumpd  - Daemon that logs packets outputs by the firewall to 
 *		 a tcpdump binary file.
 * 
 *    Author: Francis J. Lacoste <francis.lacoste@iNsu.COM>
 * 
 *    Copyright (C) 2000 Francis J. Lacoste, iNsu Innovations
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 * 
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 * 
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdarg.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#include <netinet/ip.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

typedef u_int32_t __u32;
typedef u_int16_t __u16;
typedef u_int8_t __u8;

#include <linux/netlink.h>
#ifndef MSG_TRUNC
#define MSG_TRUNC 0x20
#endif

#include <pcap.h>
#include <net/bpf.h>

#define PID_FILE_PATH "/var/run/fwdumpd.pid"

#define MAX_CAP_LEN 65535
#define MAX_LINE 1023

FILE *cap_file = NULL;	    /* The opened capture file */
char *cap_file_path = NULL; /* The path to the capture file */
int netlink = -1;	    /* Netlink socket FD */
int foreground_mode = 0;    /* Set to 1 when using debug mode */
int use_mark_flag = 0;	    /* Set to 1 when we should check fwmark */
__u32 mark;

struct fwmsg
{
    __u32 msg_len;
    __u32 fwmark;
    char  dev[IFNAMSIZ];
    union
    {
	struct iphdr iphdr;
	__u8  pkt[MAX_CAP_LEN];
    } ippkt;
};

void
log_err_doit( int err_sys, const char *fmt, va_list ap )
{
    char msg[MAX_LINE + 1];
    int len;
    
    vsnprintf( msg, MAX_LINE, fmt, ap );
    if ( err_sys ) 
    {
	len = strlen( msg );
	if ( len != MAX_LINE )
	    snprintf( (msg + len), MAX_LINE - len, ": %s", 
		      strerror(errno) );
    }
    len = strlen( msg );
    if ( len != MAX_LINE )
	snprintf( (msg + len), MAX_LINE - len, "\n" );
    
    if ( foreground_mode ) 
    {
	fputs( msg, stderr );
    } else {
	syslog( LOG_ERR, msg );
    }
}

void
log_err_sys( const char *fmt, ... )
{
    va_list ap;
    va_start( ap, fmt );
    log_err_doit( 1, fmt, ap );
    va_end(ap);
    exit(1);
}


void
log_err( const char *fmt, ... )
{
    va_list ap;
    va_start( ap, fmt );
    log_err_doit( 0, fmt, ap );
    va_end(ap);
    exit(1);
}


void
log_info( const char *fmt, ... )
{
    va_list ap;
    char msg[MAX_LINE + 1];
    int len;
    
    va_start( ap, fmt );
    vsnprintf( msg, MAX_LINE, fmt, ap );

    len = strlen( msg );
    if ( len != MAX_LINE )
	snprintf( (msg + len), MAX_LINE - len, "\n" );

    if ( foreground_mode ) 
    {
	fputs( msg, stderr );
    } else {
	syslog( LOG_INFO, msg );
    }
    va_end(ap);
}


void
open_log_file()
{
    struct pcap_file_header pfh;

    cap_file = fopen( cap_file_path, "w" );
    if ( ! cap_file )
	log_err_sys( "failed to open log file %s", cap_file_path );

    /* 
       set to tcpdump defaults 
       
       This is taken from snort by Martin Roesch.
     */
   pfh.magic = 0xa1b2c3d4;

   pfh.version_major = PCAP_VERSION_MAJOR;
   pfh.version_minor = PCAP_VERSION_MINOR;
   pfh.thiszone = 0;
   pfh.sigfigs = 0;
   pfh.snaplen = MAX_CAP_LEN;

   pfh.linktype = DLT_RAW;

   fwrite(&pfh, sizeof(struct pcap_file_header), 1, cap_file);
   if ( ferror(cap_file) )
       log_err_sys( "error writing header" );

   fflush( cap_file );
   if ( ferror(cap_file) )
       log_err_sys( "error writing header" );
}

void
rotate_file(int signal)
{
    log_info( "caught SIGHUP, reopening log file" );
    if ( fclose(cap_file) )
	log_err_sys( "error closing capture file" );
    
    open_log_file();
}

void
cleanup_exit(int signal)
{
    log_info( "caught terminating signal. Exiting." );

    if ( fclose(cap_file) )
	log_err_sys( "error closing capture file" );
    
    if ( close(netlink) )
	log_err_sys( "error closing netlink socket" );
    
    if ( !foreground_mode )
	unlink( PID_FILE_PATH );
    
    exit(0);
}

void 
summon_daemon()
{
    FILE *pidfile;
    
    switch (fork())
    {
    case -1:
	log_err_sys( "error forking" );
    case 0:
	/* Child */
	break;
    default:
	/* Parent */
	exit(0);
    }

    /* Detach from controlling terminal and become  */
    setsid();
    
    /* Close STDIN, STDOUT and STDERR and redirect to /dev/null */
    close(0);
    close(1);
    close(2);
    open( "/dev/null", O_RDWR );
    dup(0);
    dup(0);

    /* Save pid to file */
    pidfile = fopen( PID_FILE_PATH, "w" );
    if ( pidfile != NULL ) 
    {
	fprintf( pidfile, "%d\n", getpid() );
	fclose( pidfile );
    }
}

void 
open_netlink()
{
    struct sockaddr_nl snl;

    netlink = socket( PF_NETLINK, SOCK_RAW, NETLINK_FIREWALL );
    if ( netlink < 0 ) 
	log_err_sys( "socket failed" );

    /* Bind the socket to the netlink structure for anything. */
   memset( &snl, 0, sizeof snl);
   snl.nl_family = AF_NETLINK;
   snl.nl_groups = -1;

   if (bind ( netlink, (struct sockaddr *) &snl, sizeof snl) )
       log_err_sys( "bind failed" );
}

void
dump_packets()
{
    while (1)
    {
	struct fwmsg fwmsg;

	int status;
	struct iovec iov = { &fwmsg, sizeof (struct fwmsg) };
	struct sockaddr_nl snl;
	struct msghdr msg = { (void*)&snl, sizeof snl, &iov, 1, NULL, 0, 0};
	int caplen;
	struct pcap_pkthdr pcap_hdr;

	status = recvmsg ( netlink, &msg, 0);

	if (status < 0)
	{
	    if (errno == EINTR)
		continue;
	    log_err_sys( "netlink recvmsg overrun" );
	}

	if (status == 0)
	    log_err ( "netlink EOF. Exiting." );

	if (msg.msg_namelen != sizeof snl)
	    log_err( "netlink sender address length error: length %d",
		     msg.msg_namelen);

	if (msg.msg_flags & MSG_TRUNC)
	{
	    log_info ( "netlink error: message truncated" );
	    continue;
	}

	/* Got netlink packet */
	caplen = fwmsg.msg_len - (sizeof( __u32) * 2 + IFNAMSIZ);
	if( caplen <= 0 ) 
	    log_err( "invalid payload length: %d! Exiting.", caplen );

	/* Check if we should log this packet */
	if ( use_mark_flag && fwmsg.fwmark != mark )
	    continue;
	
	/* Fill pcap structure */
	if ( gettimeofday( &pcap_hdr.ts, NULL ) == -1)
	    log_err_sys( "error getting timestamp" );

	pcap_hdr.caplen = caplen;
	pcap_hdr.len = ntohs(fwmsg.ippkt.iphdr.tot_len);
	
	fwrite( &pcap_hdr, sizeof pcap_hdr, 1, cap_file);
	if ( ferror( cap_file ) )
	    log_err_sys( "error writing packet header to capture file" );
	
	fwrite( &fwmsg.ippkt, caplen, 1, cap_file );
	if ( ferror( cap_file ) )
	    log_err_sys( "error writing packet to capture file" );

	fflush(cap_file);
	if ( ferror( cap_file ) )
	    log_err_sys( "error writing packet to capture file" );
    }

}

void
usage(int status)
{
    fprintf( stderr, "Usage: fwdumpd [ -f ] [ -m <mark> ] logfile\n" 
	             "   or: fwdumpd [-hv]\n"
	     );
    fprintf( stderr, "Dump packets output by the firewall to a tcpdump"
	             " binary capture file.\n\n"
	             "  -f        stay in foreground. (For debugging).\n"
	             "  -m <mark> only logs packets with mark.\n" 
	             "  -h	  print this help message\n" 
	             "  -v        output version and copyright.\n" 
	     );
    exit(status);
}

void
version()
{
    fprintf( stderr, "fwdumpd " FWDUMPD_VERSION "\n"
	             "Copyright (c) 2000 Francis J. Lacoste, iNsu Innvations Inc.\n"
	             "Report bugs to <bugs@iNsu.COM>\n\n"
	             "There is NO warranty.  You may redistribute this software\n"
	             "under the terms of the GNU General Public License.\n"
	             "For more information about these matters, see "
	             "the file named COPYING.\n" );
    exit(0);
}

void
parse_cmdline( int argc, char *argv[] )
{
    int c;
    char *tailptr;
    
    /* Do not print error message automatically */
    opterr = 0;
    while ( (c = getopt( argc, argv, "vhfm:" )) != -1 )
    {
	switch(c)
	{
	case 'f':
	    foreground_mode = 1;
	    break;
	case 'm':
	    use_mark_flag = 1;
	    mark = (__u32)strtol( optarg, &tailptr, 10 );
	    if ( *tailptr != '\0' || mark < 0 || mark > UINT_MAX )
	    {
		fprintf( stderr, "%s: invalid fwmark\n", argv[0] );
		exit(1);
	    }
	    break;
	case 'v':
	    version();
	    /* Not reached */
	    break;
	case 'h':
	    usage(0);
	    /* Not reached */
	    break;
	case '?':
	default:
	    fprintf ( stderr, "%s: invalid option %c\n", argv[0],
		      optopt );
	    usage(1);
	    /* Not reached */
	    break;
	}
    }
    if ( optind != (argc - 1) ) 
    {
	fprintf( stderr, "%s: invalid log file argument.\n", argv[0]
		 );
	usage(1);
    }
    
    cap_file_path = strdup( argv[optind] );
}

int
main ( int argc, char *argv[] )
{
    struct sigaction sig;
    
    parse_cmdline( argc, argv );

    if ( ! foreground_mode )
	openlog( "fwdump", LOG_PID, LOG_DAEMON );

    memset( &sig, 0, sizeof(sig));

    signal( SIGTERM, cleanup_exit );
    signal( SIGINT, cleanup_exit );
    signal( SIGHUP, rotate_file );

    umask( 027 );
    open_log_file();
    open_netlink();
    if ( !foreground_mode )
	summon_daemon();

    log_info( "started and listening for packets" );
    dump_packets();

    /* Not reached */
    return 0;
}



